# Why does this file exist, and why not put this in `__main__`?
#
# You might be tempted to import things from `__main__` later,
# but that will cause problems: the code will get executed twice:
#
# - When you run `python -m duty` python will execute
#   `__main__.py` as a script. That means there won't be any
#   `duty.__main__` in `sys.modules`.
# - When you import `__main__` it will get executed again (as a module) because
#   there's no `duty.__main__` in `sys.modules`.

"""Module that contains the command line application."""

import argparse
import sys
from importlib import util as importlib_util
from typing import Dict, List, Optional, Tuple

from duty import logic

SelectionType = List[Tuple[logic.Duty, Dict[str, str]]]


def load_duties(path: str) -> Dict[str, logic.Duty]:
    """
    Load duties from a Python file.

    Arguments:
        path: The path to the Python file to load.

    Returns:
        The loaded duties.
    """
    logic.duties.clear()
    spec = importlib_util.spec_from_file_location("duty.loaded", path)
    duties = importlib_util.module_from_spec(spec)
    spec.loader.exec_module(duties)  # type: ignore
    return logic.duties


def parse_selection(args: List[str], duties: Dict[str, logic.Duty]) -> SelectionType:
    """
    Parse the arguments given on the command line to return a selection of actual duties.

    Arguments:
        args: Arguments given on the command line.
        duties: The user registered duties.

    Raises:
        ValueError: When the arguments are incorrect.

    Returns:
        A selection of duties and their keyword arguments.
    """
    selection = []
    duty: logic.Duty = None  # type: ignore
    for arg in args:
        if "=" in arg:
            if not duty:
                raise ValueError(f"argument without a duty: {arg}")
            duty_args.update((arg.split("=", 1),))  # type: ignore  # this is correct
        else:
            if duty:  # noqa: WPS513 (cannot replace by elif)
                selection.append((duty, duty_args))
            duty_args: Dict[str, str] = {}
            duty = duties[arg]
    selection.append((duty, duty_args))
    return selection


def get_parser() -> argparse.ArgumentParser:
    """
    Return the CLI argument parser.

    Returns:
        An argparse parser.
    """
    parser = argparse.ArgumentParser(prog="duty")
    parser.add_argument(
        "-d",
        "--duties-file",
        nargs=1,
        default="duties.py",
        help="Python file where the duties are defined.",
    )
    parser.add_argument("DUTIES", metavar="DUTY", nargs="+")
    return parser


def main(args: Optional[List[str]] = None) -> int:
    """
    Run the main program.

    This function is executed when you type `duty` or `python -m duty`.

    Arguments:
        args: Arguments passed from the command line.

    Returns:
        An exit code.
    """
    parser = get_parser()
    opts = parser.parse_args(args=args)

    duties = load_duties(opts.duties_file)

    try:
        selection = parse_selection(opts.DUTIES, duties)
    except KeyError as error:
        print(f"Unknown duty: {error}", file=sys.stderr)  # noqa: WPS421 (print)
        return 1
    except ValueError as error:
        print(f"Incorrect arguments: {error}", file=sys.stderr)  # noqa: WPS421 (print)
        return 1

    for duty, duty_args in selection:
        try:
            duty.run(**duty_args)
        except logic.DutyFailure as failure:
            return failure.code

    return 0



"""Module containing all the logic."""

import inspect
from typing import Any, Callable, Dict, List, Optional, Union

from failprint.cli import run as failprint_run

DutyListType = List[Union[str, Callable, "Duty"]]
CmdType = Union[str, List[str], Callable]


class DutyFailure(Exception):
    """An exception raised when a duty fails."""

    def __init__(self, code):
        """
        Initialize the object.

        Arguments:
            code: The exit code of a command.
        """
        super().__init__(self)
        self.code = code


class Duty:
    """The main duty class."""

    default_options: Dict[str, Any] = {}

    def __init__(
        self,
        name: str,
        description: str,
        function: Callable,
        pre: Optional[DutyListType] = None,
        post: Optional[DutyListType] = None,
        opts: Dict[str, Any] = None,
    ) -> None:
        """
        Initialize the object.

        Arguments:
            name: The duty name.
            description: The duty description.
            function: The duty function.
            pre: A list of duties to run before this one.
            post: A list of duties to run after this one.
            opts: Options used to create the context instance.
        """
        self.name = name
        self.description = description
        self.function = function
        self.pre = pre or []
        self.post = post or []
        self.context = Context(**(opts or self.default_options))

    def __call__(self, *args, **kwargs) -> None:
        """
        Run the duty function.

        Arguments:
            args: Positional arguments passed to the function.
            kwargs: Keyword arguments passed to the function.
        """
        self.run(*args, **kwargs)

    def run(self, *args, **kwargs) -> None:
        """
        Run the duty function.

        This function also runs pre- and post-duties.

        Arguments:
            args: Positional arguments passed to the function.
            kwargs: Keyword arguments passed to the function.
        """
        run_duties(self.pre)
        self.function(self.context, *args, **kwargs)
        run_duties(self.post)


def run_duties(duties_list: DutyListType) -> None:
    """
    Run a list of duties.

    Arguments:
        duties_list: The list of duties to run.
    """
    for duty_item in duties_list:
        if isinstance(duty_item, Duty):
            duty_item.run()
        elif isinstance(duty_item, str):
            duties[duty_item].run()
        elif callable(duty_item):
            duty_item()


class Context:
    """
    A simple context class.

    Context instances are passed to functions decorated with `duty`.
    """

    def __init__(self, **options) -> None:
        """
        Initialize the object.

        Arguments:
            options: Options passed internally to `failprint` functions.
        """
        self.options = options

    def run(self, cmd: CmdType, args=None, kwargs=None, **options):
        """
        Run a command in a subprocess or a Python callable.

        Arguments:
            cmd: A command or a Python callable.
            args: Positional arguments passed to the Python callable.
            kwargs: Keyword arguments passed to the Python callable.
            options: Options passed to `failprint` functions.

        Raises:
            DutyFailure: When the exit code / function result is greather than 0.
        """
        final_options = dict(self.options)
        final_options.update(options)
        try:
            code = failprint_run(cmd, args=args, kwargs=kwargs, **final_options)
        except KeyboardInterrupt:
            code = 130
        if code:
            raise DutyFailure(code)


duties: Dict[str, Duty] = {}


def register_duty(
    func: Callable,
    name: Optional[str] = None,
    pre: Optional[DutyListType] = None,
    post: Optional[DutyListType] = None,
    **opts,
) -> Duty:
    """
    Register a duty in the `duties` dictionary.

    Arguments:
        func: The callable to register as a duty.
        name: The duty name.
        pre: Pre-duties.
        post: Post-duties.
        opts: Options passed to the context.

    Returns:
        The registered duty.
    """
    name = name or func.__name__
    description = inspect.getdoc(func) or ""
    duties[name] = Duty(name, description, func, pre=pre, post=post, opts=opts)
    return duties[name]


def duty(*args, **kwargs) -> Union[Callable, Duty]:
    """
    Decorate a callable to transform it and register it as a duty.

    Arguments:
        args: One callable.
        kwargs: Context options.

    Raises:
        ValueError: When the decorator is misused.

    Examples:
        Decorate a function:

        ```python
        @duty
        def clean(ctx):
            ctx.run("rm -rf build", silent=True)
        ```

        Pass options to the context:

        ```python
        @duty(silent=True)
        def clean(ctx):
            ctx.run("rm -rf build")  # silent=True is implied
        ```

    Returns:
        A duty when used without parentheses, a decorator otherwise.
    """
    if args and kwargs:
        raise ValueError("The duty decorator only accepts kwargs")
    if args:
        if len(args) > 1:
            raise ValueError(
                "The duty decorator only accepts a function as first argument " "and no extra positional args",
            )
        return register_duty(args[0])

    def decorator(func):
        return register_duty(func, **kwargs)

    return decorator



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['duty']

package_data = \
{'': ['*']}

extras_require = \
{'tests': ['coverage>=5.2.1,<6.0.0',
           'invoke>=1.4.1,<2.0.0',
           'mypy>=0.782,<0.783',
           'pytest>=6.0.1,<7.0.0',
           'pytest-cov>=2.10.1,<3.0.0',
           'pytest-randomly>=3.4.1,<4.0.0',
           'pytest-sugar>=0.9.4,<0.10.0',
           'pytest-xdist>=2.1.0,<3.0.0']}

entry_points = \
{'console_scripts': ['duty = duty.cli:main']}

setup_kwargs = {
    'name': 'duty',
    'version': '0.3.0',
    'description': 'A simple task runner.',
    'long_description': '# Duty\n\n[![ci](https://github.com/pawamoy/duty/workflows/ci/badge.svg)](https://github.com/pawamoy/duty/actions?query=workflow%3Aci)\n[![documentation](https://img.shields.io/badge/docs-mkdocs%20material-blue.svg?style=flat)](https://pawamoy.github.io/duty/)\n[![pypi version](https://img.shields.io/pypi/v/duty.svg)](https://pypi.org/project/duty/)\n\nA simple task runner.\n\n## Requirements\n\nDuty requires Python 3.6 or above.\n\n<details>\n<summary>To install Python 3.6, I recommend using <a href="https://github.com/pyenv/pyenv"><code>pyenv</code></a>.</summary>\n\n```bash\n# install pyenv\ngit clone https://github.com/pyenv/pyenv ~/.pyenv\n\n# setup pyenv (you should also put these three lines in .bashrc or similar)\nexport PATH="${HOME}/.pyenv/bin:${PATH}"\nexport PYENV_ROOT="${HOME}/.pyenv"\neval "$(pyenv init -)"\n\n# install Python 3.6\npyenv install 3.6.12\n\n# make it available globally\npyenv global system 3.6.12\n```\n</details>\n\n## Installation\n\nWith `pip`:\n```bash\npython3.6 -m pip install duty\n```\n\nWith [`pipx`](https://github.com/pipxproject/pipx):\n```bash\npython3.6 -m pip install --user pipx\n\npipx install --python python3.6 duty\n```\n',
    'author': 'Timothée Mazzucotelli',
    'author_email': 'pawamoy@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pawamoy/duty',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
